/*
 * Demo how to use wv_matalloc and lapack to solve system of linear equations.
 * In Debian-like systems: install libopenblas-pthread-dev or something like that.
 * Install also liblapacke-dev for lapack.h.
 * Compile with:
 * cc -o dgesvtest dgesvtest.c wv_matalloc.c `pkg-config --libs --cflags lapack`
 * Usage:
 * ./dgesvtest [n]
 * where n is number of equations, default 1000
 * manpage dgesv: https://linux.die.net/man/l/dgesv
 * see also: http://www.netlib.org/lapack/explore-html
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <lapack.h>
#include <sys/time.h>

/* if lapack.h is not available, remove above line and uncomment the following two lines: */
//#define LAPACK_dgesv dgesv_
//void dgesv_(int *n, int *nrhs,  double *a, int *lda, int *ipiv, double *b, int *ldb, int *info);
#include "wv_matalloc.h"

double wallclock() 
{ 
   struct timeval toot;
   double r;

   gettimeofday(&toot,0);
   r=toot.tv_sec+0.000001*(double)toot.tv_usec;
   return(r);
}


int main(int argc, char *argv[])
{
   int n, nrhs, lda, ldb, info;

   n    = 1000;
   if (argc > 1)
      n = atoi(argv[1]);
   nrhs = 1;
   lda  = n+10;   // add something to lda to see if everything goes well
   ldb  = n+30;   // idem for ldb

   printf("Program dgesvtest: solve a system of linear equations\n");
   printf("using wv_matalloc and Lapack.\n");

   printf("\nNumber of equations: %d\n",n);
   /* 
    * note that the lapack and blas routines expect the matrices
    * in column-major order. In C, normally we use row-major.
    */

   double **a    = wv_matalloc(sizeof(double),0,2,n,lda);
   double  *b    = wv_matalloc(sizeof(double),0,1,ldb);
   int     *ipiv = wv_matalloc(sizeof(int),0,1,n);

   /*
    * solve for x:
    * a*x = b
    */

   int i,j;
   for (i=0; i<n; i++)
      for (j=0; j<n; j++)
	 a[i][j] = 1 - 2*drand48();
   for (i=0; i<n; i++)
      b[i] = 1 - 2*drand48();

   /*
    * a0,b0,b1 are needed to check the result.
    * a will be overwritten
    * b will contain the solution
    */
   double **a0    = wv_matalloc(sizeof(double),0,2,n,lda);
   double  *b0    = wv_matalloc(sizeof(double),0,1,ldb);
   double  *b1    = wv_matalloc(sizeof(double),0,1,ldb);

   for (i=0; i<n; i++)
      for (j=0; j<n; j++)
	 a0[i][j] = a[i][j];
   for (i=0; i<n; i++)
      b0[i] = b[i];

   double t0 = wallclock();
   LAPACK_dgesv(&n, &nrhs, &a[0][0], &lda, ipiv, &b[0], &ldb, &info);
   double t = wallclock() - t0;

   /*
    * compute the original matrix times the solution:
    * b1 = a0 * b
    * this should be equal to the original b: b0
    */
   for (i=0; i<n; i++)
   {
      b1[i] = 0;
      for (j=0; j<n; j++)
	 b1[i] += a0[j][i]*b[j];   // note the transpose usage of a0
   }

   /*
    * How to do this in fortran:
    * do i=1,n
    *    b1(i) = 0;
    *    do j=1,n
    *       b1(i) = b1(i) + a0(i,j)
    *    enddo
    * enddo
    */

   /*
    * check the result
    */
   int errors    = 0;
   double maxerr = 0;
   double sumerr = 0;
   for (i=0; i<n; i++)
   {
      if (b1[i] != b0[i])
      {
	 errors++;
	 double s = fabs(b1[i] - b0[i]);
	 sumerr += s;
	 if (s > maxerr)
	    maxerr = s;
      }
   }

   printf("\ntime (seconds): %f\n",t);

   printf("\nerrors:       %d\n",errors);
   if (errors>0)
   {
      printf("mean error: %14.6e\n",sumerr/errors);
      printf("max  error: %14.6e\n",maxerr);
   }

   printf("info:         %d",info);
   if (info == 0)
      printf(": that is OK\n");
   else
      printf(": that is NOT OK\n");
   wv_matfree(a);
   wv_matfree(b);
   wv_matfree(ipiv);
   wv_matfree(a0);
   wv_matfree(b0);
   wv_matfree(b1);
   printf("\nEnd of program dgesvtest.\n");
   return 0;
}
