/* 
 * This program compares the efficiency of the use of a wv_matalloc'ed 2-d matrix
 * compared with the efficiency of a conventional approach: 
 * an 1-d array, indexed by expressions like i*n+j.
 *
 * Indexing in first approach:
 *    a[i][j]
 * Indexing in second approach:
 *    a[i*n+j]    (n = 2nd dimension of a)
 *
 * The tests are done using a naive matrix multiplication: c = a * b
 *
 * results:
 *   gcc: Speed is in both cases almost the same
 *   icc: The wv_matalloc approach is about as fast as for gcc.
 *        The conventional approach is much faster, probably by using
 *        clever vector instructions, or by recognizing the code pattern
 *        and inserting a special template. In the case where matrix b is
 *        transposed, the results differ somewhat from the 
 *        wv_matalloc approach.
 *
 * Compilation:
 *   gcc:   gcc -Ofast -o gccmatmul matmulcompare.c wv_matalloc.c
 *   icc:   icc -Ofast -o iccmatmul matmulcompare.c wv_matalloc.c
 */
#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <math.h>
#include "wv_matalloc.h"

double wallclock() 
{ 
   struct timeval toot;
   double r;

   gettimeofday(&toot,0);
   r=toot.tv_sec+0.000001*(double)toot.tv_usec;
   return(r);
}

void matmul1(int m,int n,int k,double**a, double**b, double**c)
   // a[m,k] b[k,n] c[m,n]
{
   int i,j,l;
   for (i=0; i<m; i++)
      for (j=0; j<n; j++)
      {
	 c[i][j] = 0;
	 for (l=0; l<k; l++)
	    c[i][j] += a[i][l]*b[l][j];
      }
}

void matmul2(int m,int n,int k,double*a, double*b, double*c)
   // a[m,k] b[k,n] c[m,n]
{
   int i,j,l;
   for (i=0; i<m; i++)
      for (j=0; j<n; j++)
      {
	 c[i*n+j] = 0;
	 for (l=0; l<k; l++)
	    c[i*n+j] += a[i*k+l]*b[l*n+j];
      }
}

// note: in the following matmuls, b is transposed
void matmult1(int m,int n,int k,double**a, double**b, double**c)
   // a[m,k] b[n,k] c[m,n]
{
   int i,j,l;
   for (i=0; i<m; i++)
      for (j=0; j<n; j++)
      {
	 c[i][j] = 0;
	 for (l=0; l<k; l++)
	    c[i][j] += a[i][l]*b[j][l];
      }
}

void matmult2(int m,int n,int k,double*a, double*b, double*c)
   // a[m,k] b[n,k] c[m,n]
{
   int i,j,l;
   for (i=0; i<m; i++)
      for (j=0; j<n; j++)
      {
	 c[i*n+j] = 0;
	 for (l=0; l<k; l++)
	    c[i*n+j] += a[i*k+l]*b[j*k+l];
      }
}

void check(double *c, double *d, int m)
{
   int errors = 0;
   double sumerr = 0;
   double maxerr = 0;
   int i;
   for (i=0; i<m; i++)
      if (c[i] != d[i])
      {
	 errors++;
	 double s = fabs(c[i]-d[i]);
	 sumerr += s;
	 if (s > maxerr)
	    maxerr = s;
      }

   printf("errors: %d\n",errors);
   if (errors != 0)
   {
      printf("Mean error: %14.6e\n",sumerr/errors); 
      printf("Max  error: %14.6e\n",maxerr); 
   }
}

int main()
{
   int m,n,k;
   m = 400;
   n = 1000;
   k = 500;
   double *a  = (double*) malloc(sizeof(double)*m*k);
   double *b  = (double*) malloc(sizeof(double)*k*n);
   double *bt = (double*) malloc(sizeof(double)*n*k);
   double *c  = (double*) malloc(sizeof(double)*m*n);
   double *d  = (double*) malloc(sizeof(double)*m*n);

   double **aa  = wv_matalloc(sizeof(double),a,2,m,k);
   double **bb  = wv_matalloc(sizeof(double),b,2,k,n);
   double **bbt = wv_matalloc(sizeof(double),bt,2,n,k);
   double **cc  = wv_matalloc(sizeof(double),c,2,m,n);

   int i;
   for (i=0; i<m*k; i++)
      a[i] = drand48();
   for (i=0; i<k*n; i++)
      b[i] = drand48();
   for (i=0; i<k*n; i++)
      bt[i] = drand48();

   printf("Start of program matmulcompare\n\n");
   printf("m,n,k: %d %d %d\n",m,n,k);
   //printf("check: %f %f\n",aa[3][4],a[3*k+4]);
   //printf("check: %f %f\n",bb[3][4],b[3*n+4]);

   printf("\ncomputing c[m,n] = a[m,k] * b[k,n]\n");
   double t0;
   matmul1(m,n,k,aa,bb,cc);  // warming up the caches
   t0 = wallclock();
   matmul1(m,n,k,aa,bb,cc);
   printf("wv_matalloc : %f seconds\n",wallclock()-t0);

   matmul2(m,n,k, a, b, d);  // warming up the caches
   t0 = wallclock();
   matmul2(m,n,k, a, b, d);
   printf("conventional: %f seconds\n",wallclock()-t0);

   check(c,d,m*n);

   printf("\ncomputing c[m,n] = a[m,k] * btransposed[n,k]\n");
   matmult1(m,n,k,aa,bbt,cc);  // warming up the caches
   t0 = wallclock();
   matmult1(m,n,k,aa,bbt,cc);
   printf("wv_matalloc : %f seconds\n",wallclock()-t0);

   matmult2(m,n,k, a, bt, d);  // warming up the caches
   t0 = wallclock();
   matmult2(m,n,k, a, bt, d);
   printf("conventional: %f seconds\n",wallclock()-t0);

   check(c,d,m*n);
   printf("\nEnd of program matmulcompare\n");
}
